use clap::Parser;
use log::LevelFilter;
use std::path::PathBuf;

mod config;
mod run;

/// Arbitrage bot
#[derive(Parser, Debug)]
#[command(version, author, about)]
struct Args {
    /// Configuration file
    #[arg(short = 'c', long = "config", env = "ARBI_CONFIG")]
    config: PathBuf,

    /// Log level
    #[arg(long = "log-level", env = "ARBI_LOG_LEVEL", default_value = "info")]
    log_level: String,

    #[command(subcommand)]
    command: Command,
}

#[derive(Parser, Debug)]
enum Command {
    // Run the arbitrage bot
    Run {},

    // Check the configuration file
    CheckConfig {},
}

fn main() -> miette::Result<()> {
    let args = Args::parse();

    let log_level_filter = match args.log_level.to_lowercase().as_str() {
        "debug" => LevelFilter::Debug,
        "trace" => LevelFilter::Trace,
        "warn" => LevelFilter::Warn,
        "error" => LevelFilter::Error,
        "info" => LevelFilter::Info,
        _ => LevelFilter::Info, // Default to info
    };

    env_logger::Builder::new()
        .filter_level(log_level_filter)
        .init();

    match args.command {
        Command::Run {} => {
            let config = config::Config::parse_config(args.config)?;

            run::run(config)?;

            Ok(())
        }
        Command::CheckConfig {} => {
            let config = config::Config::parse_config(args.config)?;

            println!("Configuration correctly parsed\n{:#?}", config);

            Ok(())
        }
    }
}
