use std::{collections::HashMap, path::PathBuf, str::FromStr};

use alloy::primitives::{aliases::U112, Address};

use miette::{miette, Result};
use serde::de::{self, Visitor};
use serde::{Deserialize, Deserializer, Serialize};

use log::{debug, info};

#[derive(Debug, Serialize, Deserialize)]
pub struct Pair {
    token0: Address,
    token1: Address,
    reserve0: U112,
    reserve1: U112,
    factory: Address,
}

#[derive(Debug, Eq, Hash, PartialEq)]
struct AddressPair(Address, Address);

impl Serialize for AddressPair {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        serializer.serialize_str(format!("{:?}:{:?}", self.0, self.1).as_str())
    }
}

impl<'de> Deserialize<'de> for AddressPair {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'de>,
    {
        struct AddressPairVisitor;

        impl Visitor<'_> for AddressPairVisitor {
            type Value = AddressPair;

            fn expecting(&self, formatter: &mut std::fmt::Formatter) -> std::fmt::Result {
                formatter.write_str("a string containing two addresses separated by a colon (e.g. \"0x...0:0x...0\")")
            }

            fn visit_str<E>(self, value: &str) -> Result<Self::Value, E>
            where
                E: de::Error,
            {
                let parts: Vec<&str> = value.splitn(2, ':').map(str::trim).collect();
                if parts.len() != 2 {
                    return Err(E::custom(format!(
                        "Invalid format, expected 'address1:address2' but got \"{}\"",
                        value
                    )));
                }

                let token0 = Address::from_str(parts[0])
                    .map_err(|e| E::custom(format!("Error parsing the address \"{}\"", e)))?;
                let token1 = Address::from_str(parts[1])
                    .map_err(|e| E::custom(format!("Error parsing the address \"{}\"", e)))?;

                Ok(AddressPair(token0, token1))
            }
        }

        deserializer.deserialize_str(AddressPairVisitor)
    }
}

#[derive(Debug, Serialize, Deserialize)]
pub struct Pairs {
    pairs: HashMap<Address, Pair>,
    by_tokens: HashMap<AddressPair, Vec<Address>>,
}

impl Pairs {
    pub fn new(filename: &PathBuf) -> Result<Self> {
        debug!("Looking for pairs file in {:?}", filename);

        let path = std::path::Path::new(&filename);
        if path.exists() {
            info!("Found existing {:?}", filename);

            let data = std::fs::read_to_string(path).map_err(|e| miette!(e))?;
            serde_json::from_str(&data).map_err(|e| miette!(e))
        } else {
            info!(
                "File {:?} doesn't exist, creating new structure from scratch",
                filename
            );

            Ok(Pairs {
                pairs: HashMap::new(),
                by_tokens: HashMap::new(),
            })
        }
    }

    fn get(&self, address: Address) -> Option<&Pair> {
        self.pairs.get(&address)
    }

    pub fn add(
        &mut self,
        pair: Address,
        token0: Address,
        token1: Address,
        reserve0: U112,
        reserve1: U112,
        factory: Address,
    ) -> Option<()> {
        let old = self.pairs.insert(
            pair,
            Pair {
                token0,
                token1,
                reserve0,
                reserve1,
                factory,
            },
        );

        match old {
            Some(old) => {
                info!("Pair {} already present, updating reserves (reserve0: {} -> {}; reserve1: {} -> {})",
                    pair, old.reserve0, reserve0, old.reserve1, reserve1
                );

                Some(())
            }
            None => {
                info!("First time seeing pair {}, adding it", { pair });

                match self.by_tokens.get_mut(&AddressPair(token0, token1)) {
                    Some(tokens) => {
                        tokens.push(pair);
                        info!(
                            "Already know {} pairs with tokens {:?} and {:?}",
                            tokens.len(),
                            token0,
                            token1
                        );
                    }
                    None => {
                        self.by_tokens
                            .insert(AddressPair(token0, token1), vec![pair]);
                        info!(
                            "It's the first pair with tokens {:?} and {:?}",
                            token0, token1
                        );
                    }
                }

                None
            }
        }
    }

    pub fn dump(&self, filename: &PathBuf) -> Result<()> {
        let data = serde_json::to_string(&self).map_err(|e| miette!(e))?;
        std::fs::write(filename, data).map_err(|e| miette!(e))?;

        Ok(())
    }
}
