#[path = "pairs.rs"]
mod pairs;
#[path = "priority_queue.rs"]
mod priority_queue;

use std::{
    sync::{Arc, Mutex},
    time::Duration,
};

use crate::config::Config;
use alloy::{
    eips::BlockNumberOrTag,
    primitives::Address,
    providers::{
        fillers::FillProvider, DynProvider, Provider, ProviderBuilder, RootProvider, WsConnect,
    },
    pubsub::PubSubFrontend,
    rpc::types::Filter,
};
use futures_util::{stream, StreamExt};
use miette::{miette, Result};

use log::{debug, info};
use pairs::Pairs;
use priority_queue::{Action, PriorityQueue};
use tokio::time::sleep;

alloy::sol!(
    #[allow(missing_docs)]
    #[sol(rpc)]
    IUniswapV2Pair,
    "abi/IUniswapV2Pair.json"
);

async fn process_swaps(
    ws: WsConnect,
    priority_queue: Arc<Mutex<PriorityQueue>>,
) -> eyre::Result<()> {
    let provider = ProviderBuilder::new().on_ws(ws).await.unwrap();

    let filter = Filter::new()
        .event("Swap(address,uint256,uint256,uint256,uint256,address)")
        .from_block(BlockNumberOrTag::Latest);

    let sub = provider.subscribe_logs(&filter).await?;
    let mut stream = sub.into_stream();

    let mut latest_block = 0;

    while let Some(log) = stream.next().await {
        let block_number = log.block_number.unwrap();
        if block_number > latest_block {
            latest_block = block_number;
            info!("Processing block number {:?}", block_number);
        }

        priority_queue
            .lock()
            .unwrap()
            .push(Action::ProcessPair(log.address()));
    }

    Ok(())
}

async fn process_pair(
    ws: WsConnect,
    pairs: Arc<Mutex<Pairs>>,
    pair_address: Address,
) -> eyre::Result<()> {
    let provider = ProviderBuilder::new().on_ws(ws).await.unwrap();

    let pair = IUniswapV2Pair::new(pair_address, provider);
    let token0 = pair.token0().call().await?._0;
    let token1 = pair.token1().call().await?._0;
    let reserve0 = pair.getReserves().call().await?.reserve0;
    let reserve1 = pair.getReserves().call().await?.reserve1;
    let factory = pair.factory().call().await?._0;

    pairs
        .lock()
        .unwrap()
        .add(pair_address, token0, token1, reserve0, reserve1, factory);

    Ok(())
}

async fn consume_priority_queue(
    ws: WsConnect,
    pairs: Arc<Mutex<Pairs>>,
    priority_queue: Arc<Mutex<PriorityQueue>>,
    config: Config,
) {
    let mut guard = priority_queue.lock().unwrap();
    let actions: Vec<Action> = guard.0.drain(..).collect(); //move all actions to temporary vector in order to unlock the mutex
    drop(guard); //release before the expensive operation

    stream::iter(actions)
        .for_each_concurrent(config.concurrency, |action| {
            let pairs_clone = pairs.clone();
            let ws = ws.clone();
            async move {
                match action {
                    Action::ProcessPair(pair_address) => {
                        info!("Processing pair: {:?}", pair_address);
                        process_pair(ws, pairs_clone, pair_address).await;
                    }
                }
            }
        })
        .await;
}

async fn manage_interruption(pairs: Arc<Mutex<Pairs>>, config: Config) -> eyre::Result<()> {
    tokio::signal::ctrl_c()
        .await
        .expect("Error receiving the signal");
    info!("Program correctly interrupted by the user");

    match pairs.lock().unwrap().dump(&config.pairs_file) {
        Ok(_) => info!("Pairs correctly dumped to {:?}", &config.pairs_file),
        Err(e) => info!("Error dumping pairs to {:?} {:?}", &config.pairs_file, e),
    }

    std::process::exit(0);
}

pub fn run(config: Config) -> Result<()> {
    let runtime = tokio::runtime::Runtime::new().unwrap();
    let pairs = Arc::new(Mutex::new(Pairs::new(&config.pairs_file)?));
    let priority_queue = Arc::new(Mutex::new(PriorityQueue::new()));
    let ws = WsConnect::new(&config.endpoint);

    runtime.block_on(async {
        tokio::spawn(manage_interruption(pairs.clone(), config.clone()));

        // process all the `Swap` events adding actions to the queue
        tokio::spawn(process_swaps(ws.clone(), priority_queue.clone()));

        loop {
            consume_priority_queue(ws.clone(), pairs.clone(), priority_queue.clone(), config.clone()).await;

            debug!("The entire queue has been processed, waiting 100ms before checking if new actions are available...");
            sleep(Duration::from_millis(100)).await;
        }
    });

    Ok(())
}
